/* ITU G.722 3rd Edition (2012-09) */

/* ITU-T G722 PLC Appendix IV - Reference C code for fixed-point implementation */
/* Version:       1.2                                                           */
/* Revision Date: Jul 3, 2007                                                   */

/*
 * ITU-T G.722 PLC Appendix IV   ANSI-C Source Code
 * Copyright (c) 2006-2007
 * France Telecom
 */

#include <stdlib.h>
#include <assert.h>

#include "scratch.h"
#include "g722.h"

#include "g722_plc.h"
#include "g722_com.h"
#include "funcg722.h"
#include "mem_func.h"

#define round round_fx
Flag Overflow = 0;

/**********************************
* declaration of PLC subroutines *
**********************************/

/* lower-band analysis (main subroutine: G722PLC_ana) */
static void G722PLC_ana(G722PLC_STATE *state, g722_state *decoder);

static inline Word32 G722PLC_hp50(Word16 *x1, Word32 *y1, Word16 signal);

/* static Word16  G722PLC_pitch_ol(Word16 * signal, Word16 *maxco, Word16* overfl_shft); */
static Word16 G722PLC_pitch_ol(G722PLC_STATE *state, Word16 *maxco, Word16 *overfl_shft);

/* static Word16  G722PLC_classif_modif(Word16 maxco, Word16 nbl, Word16 nbh, Word32* mem_speech, Word16 l_mem_speech,
 * Word16* mem_exc, Word16 l_exc, Word16* t0); */
static Word16 G722PLC_classif_modif(Word16 maxco, Word16 nbl, Word16 nbh, G722PLC_STATE *state);

/* static void    G722PLC_lag_window(Word16 * R_h, Word16 * R_l, Word16 ord);
 * static void    G722PLC_levinson(Word16 R_h[], Word16 R_l[], Word16 rc[], Word16 * stable, Word16 ord, Word16 * a);
 * static void    G722PLC_autocorr(Word16 * x, Word16 * R_h, Word16 * R_l, Word16 ord, Word16 len); */
static void G722PLC_lag_window(Word32 *R, Word16 ord);

static void G722PLC_levinson(Word32 R[], Word16 rc[], Word16 *stable, Word16 ord, Word32x *a);

static void G722PLC_autocorr(Word32 *x, Word32 *R, Word16 ord, Word16 len, Word16 mode, G722PLC_STATE *state);

static void G722PLC_lpc(G722PLC_STATE *state);

static void G722PLC_residu(G722PLC_STATE *state);

/* lower-band synthesis (main subroutine: G722PLC_syn) */

/* static void   G722PLC_syn(G722PLC_STATE * state, Word32 * syn, Word16 NumSamples); */
static void G722PLC_syn(G722PLC_STATE *state, Word32 *syn, Word16 NumSamples, Word16 cyclic_mode);

/* static inline Word16  G722PLC_ltp_pred_1s(Word16* exc, Word16 t0, Word16 *jitter); */
static void G722PLC_ltp_syn(G722PLC_STATE *state, Word16 *cur_exc, Word16 *cur_syn, Word16 n, Word16 *jitter);

static inline void G722PLC_syn_filt(Word16 m, Word32x *a, Word16 *x, Word16 *y);

static inline void G722PLC_syn_filt_circ(Word16 m, Word32x *a, Word16 *x, Word16 *y, G722PLC_STATE *state);

static void G722PLC_attenuate(G722PLC_STATE *state, Word16 *cur_sig, Word16 *tabout, Word16 NumSamples,
                              Word16x *ind, Word16 *weight, Word16 mem_ind) property(loop_levels_1);

/* static void   G722PLC_attenuate_lin(G722PLC_STATE * state, Word16 fact, Word16 * cur_sig, Word16 * tabout, Word16
 * NumSamples,
 * Word16 * ind, Word16 * weight, Word16 mem_ind) property(loop_levels_1);
 * static inline void    G722PLC_calc_weight(Word16 *ind_weight, Word16 fact1, Word16 fact2p, Word16 fact3p, Word16 *
 * weight) ; */
static void G722PLC_update_mem_exc(G722PLC_STATE *state, Word16 *cur_sig, Word16 NumSamples);

/* higher-band synthesis */
static Word16 * G722PLC_syn_hb(G722PLC_STATE *state);

/***********************************
* definition of main PLC routines *
***********************************/

/*----------------------------------------------------------------------
 * G722PLC_init(l_frame)
 * allocate memory and return PLC state variables
 *
 * l_frame (i) : frame length @ 8kHz
 *---------------------------------------------------------------------- */
void * G722PLC_init(Word16 l_frame)
{
    Word16 i;
    G722PLC_STATE *restrict state;

    scratchLog(0x020000);

    /* allocate memory for PLC state */
    state = (G722PLC_STATE *)malloc_(sizeof(G722PLC_STATE));
    if (state == NULL)
    {
        exit(-1);
    }

    /* constants */
    state->l_frame = l_frame;

    /* t0SavePlus, l_exc and l_mem_speech replaced by constant preprocessor defines
     * state-> t0SavePlus = (Word16) 2;
     * l_exc and l_mem_speech replaced by constant preprocessor defines
     * state-> l_exc = add(shl(MAXPIT, 1), 1); / * 2 periods +1 for smoothing* /
     * state-> l_mem_speech = add(L_EXC, ORD_LPC); */

    /* bad frame indicator */
    state->prev_bfi = 0;

    /* LPC, pitch, signal classification parameters */
    state->a = (Word32x *)calloc_(ORD_LPC + 1, sizeof(Word32x));
    if (state->a == NULL)
    {
        exit(-1);
    }

    state->mem_syn = (Word16 *)calloc_(ORD_LPC, sizeof(Word16));
    if (state->mem_syn == NULL)
    {
        exit(-1);
    }

    for (i = 0; i < ORD_LPC; i++)
    {
        state->mem_syn[i] = (Word16)0;
        state->a[i] = (Word32)0;
    }
    state->a[ORD_LPC] = (Word32)0;
    state->t0 = (Word16)0;
    state->clas = G722PLC_WEAKLY_VOICED;

    /* signal buffers */
    state->mem_speech = (Word32 *)calloc_(L_MEM_SPEECH, sizeof(Word32));
    if (state->mem_speech == NULL)
    {
        exit(-1);
    }
    state->mem_speech_hb = (Word32 *)calloc_(LEN_HB_MEM, sizeof(Word32));    /*MAXPIT is needed, for complexity reason;
                                                                              * LEN_HB_MEM: framelength 20ms*/
    if (state->mem_speech_hb == NULL)
    {
        exit(-1);
    }
    state->mem_pitch_hb  = (Word16 *)malloc_(8 * CROSSFADELEN);
    if (state->mem_pitch_hb == NULL)
    {
        exit(-1);
    }

    /* state->mem_speech_hb = (Word16x *)calloc_(LEN_HB_MEM, sizeof(Word16x)); / *MAXPIT is needed, for complexity
     * reason; LEN_HB_MEM: framelength 20ms* / */
    state->mem_exc = (Word32 *)calloc_(L_EXC, sizeof(Word32));
    if (state->mem_exc == NULL)
    {
        exit(-1);
    }
    state->b_mem_speech =    state->mem_speech;
    state->b_mem_speech_hb = state->mem_speech_hb;
    state->b_mem_pitch_hb  = state->mem_pitch_hb;

    /* cross-fading */
    state->count_crossfade = CROSSFADELEN;

    /* higher-band hig-pass filtering */
    state->mem_hpf_in = 0;
    state->mem_hpf_out = 0;
    state->count_hpf = 32767;    /*max value*/

    /* adaptive muting */
    state->count_att = (Word16)0;
    state->count_att_hb = (Word16)0;
    state->weight_lb = 32767 << 16;
    state->weight_hb = 32767 << 16;
    state->limit_att_hb = 160;
    state->fact1 = FACT1_V;
    state->fact2 = FACT2_V;
    state->fact3 = FACT3_V;

    /* state->fact2p = FACT2P_V;   no longer needed as the absolute factors are used instead
     * state->fact3p = FACT3P_V;   no longer needed as the absolute factors are used instead */

    scratchLog(0x020001);
    return ((void *)state);
}

/*----------------------------------------------------------------------
 * G722PLC_conceal(plc_state, xl, xh, outcode, decoder)
 * extrapolation of missing frame
 *
 * plc_state (i/o) : state variables of PLC
 * xl            (o)	 : decoded lower-band
 * xh            (o)	 : decoder higher-band
 * outcode	 (o)	 : decoded synthesis
 * decoder	 (i/o) : g722 states (QMF, ADPCM)
 *---------------------------------------------------------------------- */
Word32 G722PLC_conceal(void *plc_state, short *outcode, g722_state *decoder)
{
    G722PLC_STATE *restrict state = (G722PLC_STATE *)plc_state;
    Word16 Temp;  /*, len; */
    Word32 *xl;
    Word16 *xh;

    scratchLog(0x020002);

    /***********************
    * reset counter			 *
    ***********************/

    state->count_crossfade = 0; /* reset counter for cross-fading */
    int copy_len = min(CROSSFADELEN, state->l_frame);

    /***********************
    * generate lower band *
    ***********************/

    /* check if first missing frame (i.e. if previous frame received)
     * first missing frame -> analyze past buffer + PLC
     * otherwise					 -> PLC
     */

    /* len = sub(L_MEM_SPEECH, state->l_frame);
     * xl = &state->mem_speech[len]; */
    xl = state->mem_speech;

    if (state->prev_bfi == 0)
    {
        scratchLog(0x020003);

        state->count_att = 0; /* reset counter for attenuation in lower band */
        state->count_att_hb = 0; /* reset counter for attenuation in higher band */

        /* state->weight_lb = 32767;
         * state->weight_hb = 32767; */
        state->weight_lb = 32767 << 16;
        state->weight_hb = 32767 << 16;

        /**********************************
        * analyze buffer of past samples *
        * - LPC analysis
        * - pitch estimation
        * - signal classification
        **********************************/

        G722PLC_ana(state, decoder);

        /******************************
        * synthesize missing samples *
        ******************************/

        /* set increment for attenuation */
        if (state->clas == G722PLC_VUV_TRANSITION)
        {
            scratchLog(0x020004);

            /* attenuation in 20 ms */

            /* printf("20ms\n"); */
            state->limit_att_hb = 80;
            state->fact1 = FACT1_UV;

            /* state->fact2p = FACT2P_UV;   no longer needed as the absolute factors are used instead
             * state->fact3p = FACT3P_UV;   no longer needed as the absolute factors are used instead */
            state->fact2 = FACT2_UV;
            state->fact3 = FACT3_UV;

            /* Temp = FACT3_UV; */
        }
        else
        {
            /* attenuation in 40 ms */

            /* printf("40ms\n"); */
            scratchLog(0x020005);

            state->limit_att_hb = 160;
            state->fact1 = FACT1_V;

            /* state->fact2p = FACT2P_V;   no longer needed as the absolute factors are used instead
             * state->fact3p = FACT3P_V;   no longer needed as the absolute factors are used instead */
            state->fact2 = FACT2_V;
            state->fact3 = FACT3_V;

            /* Temp = FACT2_V; */
        }

        scratchLog(0x020006);

        if (state->clas == G722PLC_TRANSIENT)
        {
            /* printf("10ms\n"); */

            /* attenuation in 10 ms */
            state->limit_att_hb = 40;
            state->fact1 = FACT1_V_R;

            /* state->fact2p = FACT2P_V_R;   no longer needed as the absolute factors are used instead
             * state->fact3p = FACT3P_V_R;   no longer needed as the absolute factors are used instead */
            state->fact2 = FACT2_V_R;
            state->fact3 = FACT3_V_R;

            /* Temp = FACT1_V_R; */
        }

        scratchLog(0x020007);

        /* synthesize lost frame, high band */
        xh = G722PLC_syn_hb(state);

        /*shift low band*/

        /* G722PLC_copy(&state->mem_speech[state->l_frame],state->mem_speech,len); / *shift low band* / */
        state->mem_speech = cyclic_add(state->mem_speech, state->l_frame, state->b_mem_speech, L_MEM_SPEECH); /*shift
                                                                                                               * low
                                                                                                               * band*/

        /* synthesize lost frame, low band directly to state->mem_speech*/
        G722PLC_syn(state, xl, state->l_frame, 1);

        /* synthesize cross-fade buffer (part of future frame)*/
        G722PLC_syn(state, state->crossfade_buf, CROSSFADELEN, 0);

        scratchLog(0x020008);

        /* attenuate outputs */

        /* if (state->l_frame <= 80) / * mode <= 10 ms, first lost frame : linear weighting* /
         * {
         * G722PLC_attenuate_lin(state, state->fact1, xl, xl, state->l_frame, &state->count_att, &state->weight_lb, LB);
         * if(sub(state->clas, G722PLC_TRANSIENT) == 0)
         * if((state->clas == G722PLC_TRANSIENT) && (state->count_att > 80))
         * {
         * state->weight_lb = 0;
         * }
         * G722PLC_attenuate_lin(state, Temp, state->crossfade_buf, state->crossfade_buf, CROSSFADELEN,
         * &state->count_att, &state->weight_lb, XF);
         * G722PLC_attenuate_lin(state, state->fact1, xh, xh, state->l_frame, &state->count_att_hb, &state->weight_hb,
         * HB);
         * }
         * else
         * { */
        scratchLog(0x020009);

        G722PLC_attenuate(state, xl, xl, state->l_frame, &state->count_att, &state->weight_lb, LB);
        G722PLC_attenuate(state,
                          state->crossfade_buf,
                          state->crossfade_buf,
                          CROSSFADELEN,
                          &state->count_att,
                          &state->weight_lb,
                          XF);
        G722PLC_attenuate(state, xh, xh, state->l_frame, &state->count_att_hb, &state->weight_hb, HB);

        /* } */
        scratchLog(0x02000A);
    }
    else
    {
        scratchLog(0x02000B);

        /* G722PLC_copy(&state->mem_speech[state->l_frame],state->mem_speech,len); / *shift* / */
        state->mem_speech = cyclic_add(state->mem_speech, state->l_frame, state->b_mem_speech, L_MEM_SPEECH); /*shift
                                                                                                               * low
                                                                                                               * band*/

        /* / * copy samples from cross-fading buffer (already generated in previous bad frame decoding)  * /
         * G722PLC_copy(state->crossfade_buf, xl, CROSSFADELEN); */
        Word16 *ptr_xl, *ptr_xf;
        ptr_xl = xl;
        for (int i = 0; i < copy_len; i++)
        {
            *ptr_xl = state->crossfade_buf[i];
            ptr_xl = cyclic_add(ptr_xl, 1, state->b_mem_speech, L_MEM_SPEECH);
        }

        /* synthesize rest of lost frame */

        /* Temp = sub(state->l_frame, CROSSFADELEN); */
        Temp = state->l_frame - CROSSFADELEN;
        if (Temp > 0)
        {
            G722PLC_syn(state, &xl[CROSSFADELEN], Temp, 0); /* remaining synthese for lost frame*/
            G722PLC_attenuate(state,
                              &xl[CROSSFADELEN],
                              &xl[CROSSFADELEN],
                              Temp,
                              &state->count_att,
                              &state->weight_lb,
                              LB);
        }
        else if (Temp < 0)
        {
            for (int i = copy_len; i < CROSSFADELEN; i++)
            {
                state->crossfade_buf[i - copy_len] = state->crossfade_buf[i];
            }
        }

        scratchLog(0x02000D);

        /* synthesize cross-fade buffer (part of future frame) and attenuate output */
        G722PLC_syn(state, &state->crossfade_buf[CROSSFADELEN - copy_len], copy_len, 0);
        G722PLC_attenuate(state,
                          &state->crossfade_buf[CROSSFADELEN - copy_len],
                          &state->crossfade_buf[CROSSFADELEN - copy_len],
                          copy_len,
                          &state->count_att,
                          &state->weight_lb,
                          XF);

        scratchLog(0x02000E);

        xh = G722PLC_syn_hb(state);
        G722PLC_attenuate(state, xh, xh, state->l_frame, &state->count_att_hb, &state->weight_hb, HB);

        scratchLog(0x02000F);
    }

    /************************
    * generate higher band *
    ************************/

    /* reset DC-remove filter states for good frame decoding */
    if (state->count_hpf >= 32000) /* 4 s good frame HP filtering (offset remove) */
    {
        state->mem_hpf_in = 0;              /* memory of higher band (input of DC-remove filter) */
        state->mem_hpf_out = 0;             /* memory of higher band (output of DC-remove filter) */
    }
    state->count_hpf = 0;               /* reset counter used to activate DC-remove filter in good frame mode  */

    /*****************************************
    * QMF synthesis filter and state update *
    *****************************************/

    G722PLC_qmf_updstat(outcode, state->l_frame, decoder, xl, xh, plc_state);

    state->prev_bfi++;

    scratchLog(0x0200FF);

    return (state->l_frame << 1);
}

/*----------------------------------------------------------------------
 * G722PLC_copy(in, out, n)
 * buffer copy
 *
 * in  (i) : input buffer
 * out (o) : output buffer
 * n	 (i) : number of samples
 *----------------------------------------------------------------------*/

void G722PLC_copy(Word32 *in, Word32 *out, Word16 n)
{
    Word16 i;
    for (i = 0; i < n; i++)
    {
        out[i] = *in++;
    }
}

/*----------------------------------------------------------------------
 * G722PLC_clear(plc_state)
 * free memory and clear PLC state variables
 *
 * plc_state (i) : PLC state variables
 *---------------------------------------------------------------------- */
void G722PLC_clear(void *plc_state)
{
    G722PLC_STATE *state = (G722PLC_STATE *)plc_state;

    free_(state->mem_speech);
    free_(state->mem_speech_hb);
    free_(state->mem_pitch_hb);
    free_(state->mem_exc);

/*	(Word32x)free_(state->a); */
    free_(state->mem_syn);
    free_(state);
}

/*********************************
* definition of PLC subroutines *
*********************************/

/*----------------------------------------------------------------------
 * G722PLC_hp50(x1, y1, signal)
 * 50 Hz high-pass filter
 *
 * x1          (i/o) : filter memory
 * y1          (i/o) : filter memory
 * signal	     (i)   : input sample
 *----------------------------------------------------------------------*/

inline Word32 G722PLC_hp50(Word16 *x1, Word32 *y1, Word16 signal)
{
    accum_t acc0;
    Word32 y1out;

    /*  y[i] =      x[i]   -         x[i-1]    */
    /*                     + 123/128*y[i-1]    */
    acc0  = to_accum((signal - *x1));
    acc0 += fract_mult(*y1, G722PLC_a_hp[1]);
    y1out = rnd_saturate(acc0);
    *y1 = y1out;
    *x1 = signal;

    /* return(round(y1out)); */
    return ((y1out));
}

/*----------------------------------------------------------------------
 * G722PLC_syn_hb(state, xh, n)
 * reconstruct higher-band by pitch prediction
 *
 * state (i/o)	 : state variables of PLC
 *---------------------------------------------------------------------- */

static Word16 * G722PLC_syn_hb(G722PLC_STATE *restrict state)
{
    Word16   *ptr;
    Word16   *ptr2;
    Word16   *ptr3;
    Word16   *xh;
    Word16 tmp_t0;
    Word16 tmp2;

    /* Word16    loop_size; */

    /* create tmp pitch delay */
    tmp_t0 = state->t0;

    /* if signal is not voiced, cut harmonic structure by forcing a 10 ms pitch */
    if (sub(state->clas, G722PLC_VOICED) != 0)
    {
        tmp_t0 = 80;
    }

/* reconstruct higher band signal by pitch prediction */

    /* tmp = state->l_frame - tmp_t0;
     * ptr = state->mem_speech_hb + sub(LEN_HB_MEM, state->t0); / *beginning of copy zone* /
     * tmp2 = LEN_HB_MEM - state->l_frame; */
    tmp2 = -state->l_frame - tmp_t0;

    /* ptr2 = state->mem_speech_hb + tmp2; / *beginning of last frame in mem_speech_hb* / */
    xh  = state->mem_speech_hb;
    state->mem_speech_hb = cyclic_add(state->mem_speech_hb, state->l_frame, state->b_mem_speech_hb, LEN_HB_MEM);    /*shift
                                                                                                                     * 1
                                                                                                                     * frame*/
    ptr = xh;
    ptr2  = cyclic_add(state->mem_speech_hb, tmp2, state->b_mem_speech_hb, LEN_HB_MEM);
    ptr3 = ptr2;

    /* loop_size = min(tmp_t0, state->l_frame); */

    /* if ((tmp <= 0)) / * l_frame <= t0; only possible in 10 ms mode* /
     * { */

    /* temporary save of new frame in state->mem_speech[0 ...state->l_frame-1] of low_band!! that will be shifted
     * after*/

    /* G722PLC_copy(ptr, state->mem_speech, state->l_frame);
     * G722PLC_copy(ptr2, state->mem_speech_hb, state->l_frame); / *shift 1 frame* /
     * G722PLC_copy(state->mem_speech, ptr2, state->l_frame);
     * }
     * else / *t0 < state->l_frame* /
     * {
     * G722PLC_copy(ptr2, state->mem_speech_hb, tmp2); / *shift memory* /
     * G722PLC_copy(ptr, ptr2, state->t0); / *copy last period* /
     * G722PLC_copy(ptr2, &ptr2[state->t0], tmp); / *repeate last period* / */
#if 0

    /*copy last period*/
    for (int i = 0; i < loop_size; i++)
    {
        *ptr = *ptr2;
        ptr  = cyclic_add(ptr, 1, state->b_mem_speech_hb, LEN_HB_MEM);
        ptr2 = cyclic_add(ptr2, 1, state->b_mem_speech_hb, LEN_HB_MEM);
    }
    if (tmp > 0) /*&&(state->l_frame==LEN_HB_MEM/2))/ * l_frame <= t0; only possible in 10 ms mode* / */
    {
        /*repeat last period*/
        for (int i = 0; i < tmp; i++)
        {
            *ptr = *ptr3;
            ptr  = cyclic_add(ptr, 1, state->b_mem_speech_hb, LEN_HB_MEM);
            ptr3 = cyclic_add(ptr3, 1, state->b_mem_speech_hb, LEN_HB_MEM);
        }
    }
#else  /* if 0 */

    /*if first bad frame copy last period into pitch buffer*/
    if (state->prev_bfi == 0)
    {
        ptr3 = state->b_mem_pitch_hb;
        state->mem_pitch_hb = state->b_mem_pitch_hb;
        for (int i = 0; i < tmp_t0; i++)
        {
            *ptr3 = *ptr2;
            ptr2 = cyclic_add(ptr2, 1, state->b_mem_speech_hb, LEN_HB_MEM);
            ptr3 = cyclic_add(ptr3, 1, state->b_mem_pitch_hb, tmp_t0);
        }
    }

    /*copy last period(s)*/
    ptr3 = state->mem_pitch_hb;
    for (int i = 0; i < state->l_frame; i++)
    {
        *ptr = *ptr3;
        ptr  = cyclic_add(ptr, 1, state->b_mem_speech_hb, LEN_HB_MEM);
        ptr3 = cyclic_add(ptr3, 1, state->b_mem_pitch_hb, tmp_t0);
    }
    state->mem_pitch_hb = ptr3;

#endif /* if 0 */
    return (xh);
}

/*-------------------------------------------------------------------------*
 * G722PLC_attenuate(state, in, out, n, count, weight)
 * linear muting with adaptive slope
 *
 * state (i/o) : PLC state variables
 * in    (i)   : input signal
 * out	 (o)   : output signal = attenuated input signal
 * n	   (i)   : number of samples
 * count (i/o) : counter
 * weight (i/o): muting factor
 *--------------------------------------------------------------------------*/

static void G722PLC_attenuate(G722PLC_STATE *restrict state,
                              Word16 *in,
                              Word16 *out,
                              Word16 n,
                              Word16x *count,
                              Word16 *weight,
                              Word16 mem_ind)
{
    Word16 i;
    Word16  *b_ptr;
    Word16 l_mem;

    /* Word16 inc_att; */
    Word16 loop_end;
    Word16 loop_start;
    Word16 fact;
    Word16 part_end;
    Word16 loc_weight;

    loc_weight = *weight;
    if (mem_ind == LB)
    {
        b_ptr = state->b_mem_speech;
        l_mem = L_MEM_SPEECH;
    }
    else if (mem_ind == XF)
    {
        b_ptr = in;
        l_mem = CROSSFADELEN;
    }
    else
    {
        b_ptr = state->b_mem_speech_hb;
        l_mem = LEN_HB_MEM;
    }
    loop_end = n;
#if 1
    i = 0;
    fact = state->fact1;

    part_end = END_1ST_PART;
    do
    {
        if (*count >= END_1ST_PART)
        {
            fact = state->fact2;
            part_end = END_2ND_PART;
        }
        if (*count >= END_2ND_PART)
        {
            fact = state->fact3;
            part_end = END_3RD_PART;
        }
        if (*count >= END_3RD_PART)
        {
            fact = state->fact3;
            part_end = 0x7FFFFFFF;
        }
        loop_end = min(n, part_end - *count);
        if (loop_end == i)
        {
            loop_end = n;
        }
        loop_start = i;
        for (; i < loop_end; i++)
            chess_loop_range(2, CROSSFADELEN)
            {
                /* calculate attenuation factor and multiply */
                loc_weight =  max(0, loc_weight - fact);
                *out = rnd_saturate(fract_mult_su(*out, loc_weight));
                out = cyclic_add(out, 1, b_ptr, l_mem);
            }
            *count = *count + loop_end - loop_start; /**loop_end; */
    }
    while (i < n);
    *weight = loc_weight;
#else  /* if 1 */
    for (i = 0; i < n; i++)
        chess_loop_range(2, CROSSFADELEN)
        {
            /* calculate attenuation factor and multiply */
            G722PLC_calc_weight(count, state->fact1, state->fact2p, state->fact3p, weight);

            /* out[i] = mult_r(*weight, in[i]); */
            *out = rnd_saturate(fract_mult(*weight, *out));

            /* *count = add(*count, state->inc_att); */
            *count = *count + state->inc_att;
            out = cyclic_add(out, 1, b_ptr, l_mem);
        }
 #endif /* if 1 */
}

/*-------------------------------------------------------------------------*
 * G722PLC_attenuate_lin(state, fact, in, out, n, count, weight)
 * linear muting with fixed slope
 *
 * state (i/o) : PLC state variables
 * fact  (i/o) : muting parameter
 * in    (i)   : input signal
 * out	 (o)   : output signal = attenuated input signal
 * n	   (i)   : number of samples
 * count (i/o) : counter
 * weight (i/o): muting factor
 *--------------------------------------------------------------------------*/
#if 0
static void G722PLC_attenuate_lin(G722PLC_STATE *state,
                                  Word16 fact,
                                  Word16 *in,
                                  Word16 *out,
                                  Word16 n,
                                  Word16 *count,
                                  Word16 *weight,
                                  Word16 mem_ind)
{
    Word16 i;
    Word16  *b_ptr, *ptr;
    Word16 l_mem;

    if (mem_ind == LB)
    {
        b_ptr = state->b_mem_speech;
        l_mem = L_MEM_SPEECH;
    }
    else if (mem_ind == XF)
    {
        b_ptr = in;
        l_mem = CROSSFADELEN;
    }
    else
    {
        b_ptr = state->b_mem_speech_hb;
        l_mem = LEN_HB_MEM;
    }
    for (i = 0; i < n; i++)
    {
        /* calculate attenuation factor and multiply */

        /* *weight = sub(*weight, fact); */
        *weight = max(*weight - fact, 0);

        /* *weight = *weight - fact;
         * out[i] = mult_r(*weight, in[i]); */
        *out = rnd_saturate(fract_mult(*weight, *out));

        /* in = cyclic_add(in, 1, state->b_mem_speech, L_MEM_SPEECH); */
        *count = *count + state->inc_att;
        out = cyclic_add(out, 1, b_ptr, l_mem);
    }

    /* *count = add(*count, i_mult(state->inc_att, 80));
     * *count = *count+(state->inc_att*n); */
}

/*-------------------------------------------------------------------------*
 * G722PLC_calc_weight(ind_weight, Ind1, Ind12, Ind123, Rlim1, Rlim2, fact1, fact2, fact3,
 *                     tab_len, WeightEnerSynthMem)
 * calculate attenuation factor
 *--------------------------------------------------------------------------*/

static inline void G722PLC_calc_weight(Word16 *ind_weight, Word16 fact1, Word16 fact2p, Word16 fact3p, Word16 *weight)
{
    /* *weight = sub(*weight, fact1);
     * if (sub(*ind_weight, END_1ST_PART) >= 0)
     * {
     * *weight = sub(*weight, fact2p);
     * }
     * if (sub(*ind_weight, END_2ND_PART) >= 0)
     * {
     * *weight = sub(*weight, fact3p);
     * }
     * if (sub(*ind_weight, END_3RD_PART) >= 0)
     * {
     * *weight = 0;
     * } */
    *weight = *weight - fact1;
    if (*ind_weight >= END_1ST_PART)
    {
        *weight = *weight - fact2p;
    }
    if (*ind_weight >= END_2ND_PART)
    {
        *weight = *weight - fact3p;
    }
    if (*ind_weight >= END_3RD_PART)
    {
        *weight = 0;
    }
    if (*weight <= 0)
    {
        *ind_weight = END_3RD_PART;
    }
    *weight = max(*weight, 0);
}

#endif /* if 0 */
/*-------------------------------------------------------------------------*
 * Function G722PLC_update_mem_exc                                          *
 * Update of state->mem_exc and shifts the memory                           *
 * if state->t0 > state->l_frame                                            *
 *--------------------------------------------------------------------------*/
static void G722PLC_update_mem_exc(G722PLC_STATE *restrict state, Word16 *exc, Word16 n)
{
    Word16   *ptr;
    Word16 temp;
    Word16 lag;

    /* shift ResMem, if t0 > l_frame */
    lag = add(state->t0, T0SAVEPLUS);    /*update temp samples*/
    temp = sub(lag, n);

    ptr = state->mem_exc + sub(L_EXC, lag);

    if (temp > 0)
    {
        G722PLC_copy(&ptr[n], ptr, temp);
        G722PLC_copy(exc, &ptr[temp], n);
    }
    else
    {
        /* copy last "pitch cycle" of residual */
        G722PLC_copy(&exc[sub(n, lag)], ptr, lag);
    }
}

/*-------------------------------------------------------------------------*
* Function G722PLC_ana(state, decoder)									                   *
* Main analysis routine
*
* state   (i/o) : PLC state variables
* decoder (i)   : G.722 decoder state variables
*-------------------------------------------------------------------------*/

static void G722PLC_ana(G722PLC_STATE *restrict state, g722_state *decoder)
{
    Word16 maxco, overfl_shft;

    /* estimate (open-loop) pitch */
    state->t0 = G722PLC_pitch_ol(state, &maxco, &overfl_shft);

    /* perform LPC analysis and compute residual signal */
    G722PLC_lpc(state);
    G722PLC_residu(state);

    /* update memory for LPC
     * during ereased period the state->mem_syn contains the non weighted
     * synthetised speech memory. For the	first erased frame, it
     * should contain the output speech.
     * Saves the last ORD_LPC samples of the output signal in
     * state->mem_syn    */

    /* G722PLC_cyclic_copy(&state->mem_speech[sub(L_MEM_SPEECH, ORD_LPC)],
     * state->mem_syn, ORD_LPC, state); */
    Word32 *in;
    Word32 *out;
    out = state->mem_syn;
    in = cyclic_add(state->mem_speech, sub(L_MEM_SPEECH, ORD_LPC), state->b_mem_speech, L_MEM_SPEECH);
    for (int i = 0; i < ORD_LPC; i++)
    {
        *out = *in;
        out++;
        in = cyclic_add(in, 1, state->b_mem_speech, L_MEM_SPEECH);
    }

    /* determine signal classification and modify residual in case of transient */
    state->clas = G722PLC_classif_modif(maxco, decoder->nbl, decoder->nbh, state);
}

/*-------------------------------------------------------*
* Function G722PLC_lag_window()																 *
*																											 *
* G722PLC_lag_window on G722PLC_autocorrelations.											 *
*																											 *
* r[i] *= lag_wind[i]																	 *
*																											 *
*	r[i] and lag_wind[i] are in special double precision.*
*	See "oper_32b.c" for the format                                          *
*																											 *
*-------------------------------------------------------*/

static void G722PLC_lag_window(Word32 *R, Word16 ord)
{
    Word32 x, xt;
    Word16 i;

    /* printf("lag_window\n"); */
    for (i = 1; i <= ord; i++)
    {
        /* x = Mpy_32(R_h[i], R_l[i], G722PLC_lag_h[i - 1], G722PLC_lag_l[i - 1]); */
        R[i] = rnd_saturate(fract_mult(R[i], G722PLC_lag[i - 1]));

        /* L_Extract(x, &R_h[i], &R_l[i]); */
    }
}

/*-------------------------------------------------------------------------*
 * Function G722PLC_levinson                                                                    *
 *--------------------------------------------------------------------------*/

static void G722PLC_levinson(Word32 R[],    /* (i)      : Rh[M+1] Vector of G722PLC_autocorrelations */

                             /* Word16 R_h[],	/ * (i)		 : Rl[M+1] Vector of G722PLC_autocorrelations (msb) * /
                              * Word16 R_l[],	/ * (i)		 : Rl[M+1] Vector of G722PLC_autocorrelations (lsb) * / */
                             Word32 rc[],                 /* (o) Q15 : rc[M]	 Reflection coefficients.
                                                           * */
                             Word16 *stable, Word16 ord, Word32x *a)
{
    Word32 t0, t1, t2;                 /* temporary variable */
    div_t tdiv; /* temporary division variable */

    /* Word16	 *A_h;								/ * LPC coef. in double prec.	* /
     * Word16	 *A_l;								/ * LPC coef. in double prec.	* /
     * Word16	 *An_h;                             / * LPC coef. for next iteration in double prec.  * /
     * Word16	 *An_l;                             / * LPC coef. for next iteration in double prec.  * / */
    Word32   *AC;                               /* LPC coef.    */
    Word32   *ACn;                          /* LPC coef. for next iteration   */

    Word16 i, j;

    /* Word16		hi, lo;
     * Word16		K_h, K_l;           / * reflexion coefficient; hi and lo * / */
    Word32 K;                  /* reflexion coefficient*/
    Word16 alp_exp;    /* Prediction gain exponent */
    Word32 alp;    /* Prediction gain mantisa */

    i = add(ord, 1);

    /* A_h = (Word16 *)calloc_(i, sizeof(Word16));
     * A_l = (Word16 *)calloc_(i, sizeof(Word16));
     * An_h = (Word16 *)calloc_(i, sizeof(Word16));
     * An_l = (Word16 *)calloc_(i, sizeof(Word16)); */
    AC = (Word32 *)malloc_(i * sizeof(Word32));
    ACn = (Word32 *)malloc_(i * sizeof(Word32));

    /* K = A[1] = -R[1] / R[0] */

    *stable = 0;

    /* t1 = L_Comp(R_h[1], R_l[1]);
     * t2 = L_abs(t1);                          / * abs R[1] * /
     * t0 = Div_32(t2, R_h[0], R_l[0]);	/ * R[1]/R[0] * / */

    /* if (t1 > 0)
     * {
     * t0 = L_negate(t0);					/ * -R[1]/R[0] * /
     * }
     * L_Extract(t0, &K_h, &K_l);		/ * K in DPF * / */

    /* t1 = R[1];
     * t2 = -R[1];
     * t0 = div32_nr_by_dr_for_q(t2, R[0], -1); */
    K = (int)((((long long)-R[1]) << 31) / ((long long)R[0]));

    /* rc[0] = K_h; */
    rc[0] = K;

    /* t0 = L_shr(K, 4);
     * L_Extract(t0, &A_h[1], &A_l[1]);	/ * A[1] in DPF * / */
    AC[1] = K >> 4;    /* A[1] */

    /*	Alpha = R[0] * (1-K**2) */

    t0 = rnd_saturate(fract_mult(K, K));    /* square K */
    /* t0 = L_sub((Word32) 0x7fffffffL, t0); / * 1 - K*K * / */
    t0 = 0x7fffffffL - t0;    /* 1 - K*K */
    t0 = rnd_saturate(fract_mult(R[0], t0));

    /* t0 = Mpy_32(K_h, K_l, K_h, K_l);	/ * K*K * /
     * t0 = L_abs(t0);                          / * Some case <0 !! * /
     * t0 = L_sub((Word32) 0x7fffffffL, t0); / * 1 - K*K * /
     * L_Extract(t0, &hi, &lo);
     * t0 = Mpy_32(R_h[0], R_l[0], hi, lo);	/ * Alpha in DPF format * / */

    /* Normalize Alpha */

    /* alp_exp = norm_l(t0); */
    alp_exp = norm(to_accum(t0));
    alp = t0 << alp_exp;

    /* t0 = L_shl(t0, alp_exp);
     * L_Extract(t0, &alp_h, &alp_l);	/ * DPF format * / */

    /*-------------------------------------- */
    /* ITERATIONS  I=2 to lpc_order */
    /*-------------------------------------- */
    accum_t t0acc;
    for (i = 2; i <= ord; i++)
    {
        /* t0 = SUM ( R[j]*A[i-j] ,j=1,i-1 ) +	R[i] */

        /* t0 = Mpy_32(R_h[1], R_l[1], A_h[i - 1], A_l[i - 1]); */
        t0acc = fract_mult(R[1], AC[i - 1]);

        /* t0 = rnd_saturate(t0acc); */
        for (j = 2; j < i; j++)
        {
            /* t0 = L_add(t0, Mpy_32(R_h[j], R_l[j], A_h[i - j], A_l[i - j])); */
            t0acc += fract_mult(R[j], AC[i - j]);

            /* t0 = rnd_saturate(t0acc); */
        }

        /* t0 = L_shl(t0, 4); */
        t0 = rnd_saturate(t0acc << 4);

        t0 = t0 + R[i];

        /* t1 = L_Comp(R_h[i], R_l[i]);
         * t0 = L_add(t0, t1);              / * add R[i] * / */

        /* / * K = -t0 / Alpha * / */

        /* t1 = L_abs(t0);
         * t2 = Div_32(t1, alp_h, alp_l);	/ * abs(t0)/Alpha * / */

        /* if (t0 > 0)
         * {
         * t2 = L_negate(t2);				/ * K =-t0/Alpha * /
         * } */
        t2 = -(int)((((long long)t0) << 31) / ((long long)alp)); /* t0/Alpha */
        t2 = t2 << alp_exp;       /* denormalize; compare to Alpha */

        /* L_Extract(t2, &K_h, &K_l);	/ * K in DPF * / */
        K = t2;

        /* rc[i - 1] = K_h; */
        rc[i - 1] = K;    /*&0xFFFF0000; */

        /* Test for unstable filter. If unstable keep old A(z) */

        /* t0 = abs(K);
         * t0 = abs(K>>16); */
        if ((abs(K >> 16) - 32750) > 0)
        {
            *stable = 1;

            /* free_(A_h);
             * free_(A_l);
             * free_(An_h);
             * free_(An_l); */
            free_(AC);
            free_(ACn);
            return;
        }

        /*------------------------------------------ */
        /*	Compute new LPC coeff. -> An[i]                  */
        /*	An[j]= A[j] + K*A[i-j]		 , j=1 to i-1  */
        /*	An[i]= K																 */
        /*------------------------------------------ */

        for (j = 1; j < i; j++)
        {
            /* t0 = Mpy_32(K_h, K_l, A_h[i - j], A_l[i - j]);
             * t0 = L_add(t0, L_Comp(A_h[j], A_l[j]));
             * L_Extract(t0, &An_h[j], &An_l[j]);
             * t0acc = fract_mult(K, AC[i-j]) to_accum(AC[j]);
             * t0acc += to_accum(AC[j]); */
            ACn[j] = rnd_saturate(fract_mult(K, AC[i - j]) + to_accum(AC[j]));
        }

        /* t2 = L_shr(t2, 4);
         * L_Extract(t2, &An_h[i], &An_l[i]); */
        ACn[i] = K >> 4;

        /*	Alpha = Alpha * (1-K**2) */

        t0 = rnd_saturate(fract_mult(K, K)); /* square K */
        /* t0 = L_abs(t0);                      / * Some case <0 !! * /
         * t0 = L_sub((Word32) 0x7fffffffL, t0); / * 1 - K*K * / */
        t0 = abs(t0);                       /* Some case <0 !! */
        t0 = 0x7fffffffL - t0; /* 1 - K*K */
        t0 = rnd_saturate(fract_mult(alp, t0));

        /* t0 = Mpy_32(K_h, K_l, K_h, K_l);	/ * K*K * /
         * t0 = L_abs(t0);                      / * Some case <0 !! * /
         * t0 = L_sub((Word32) 0x7fffffffL, t0); / * 1 - K*K * /
         * L_Extract(t0, &hi, &lo);		/ * DPF format * /
         * t0 = Mpy_32(alp_h, alp_l, hi, lo); */

        /* Normalize Alpha */

        /* j = norm_l(t0); */
        j = norm(to_accum(t0));

        /* alp = L_shl(t0, j); */
        alp = t0 << j;

        /* L_Extract(t0, &alp_h, &alp_l);	/ * DPF format * /
         * alp_exp = add(alp_exp, j);	/ * Add normalization to alp_exp * / */
        alp_exp = alp_exp + j;    /* Add normalization to alp_exp */

        /* A[j] = An[j] */

        for (j = 1; j <= i; j++)
        {
            /* A_h[j] = An_h[j];
             * A_l[j] = An_l[j]; */
            AC[j] = ACn[j];
        }
    }

    a[0] = 4096 << 16;

    for (i = 1; i <= ord; i++)
    {
        /* t0 = L_Comp(A_h[i], A_l[i]);
         * a[i] = round(L_shl(t0, 1));
         * a[i] = round(L_shl(AC[i], 1));
         * a[i] = round(AC[i]<<1)<<16; */
        a[i] = rnd_saturate(to_accum(AC[i]) << 1);
    }

    /* free_(A_h);
     * free_(A_l);
     * free_(An_h);
     * free_(An_l); */
    free_(AC);
    free_(ACn);
}

/*-------------------------------------------------------------------------*
 * Function G722PLC_autocorr																*
 *--------------------------------------------------------------------------*/

void G722PLC_autocorr(Word32 x[],    /* (i)    : Input signal                      */

                      /* Word16 r_h[],/ * (o)    : Autocorrelations  (msb)           * /
                       * Word16 r_l[], / * (o)    : Autocorrelations  (lsb)           * / */
                      Word32 r[],             /* (o)    : Autocorrelations  (lsb)           */
                      Word16 ord,             /* (i)    : LPC order                         */
                      Word16 len,             /* (i)    : length of analysis                   */
                      Word16 mode,            /* (i)    : enable circular buffer                   */
                      G722PLC_STATE *restrict state         /* (i)    : PLC state                   */
                      )
{
    Word32 sum;
    Word16 i, j, normx, tmp;
    Word32 *y = (Word32 *)malloc_(len * sizeof(Word16));
    accum_t acc;
    Word32 *ptr_sig, *ptr_base;

    /* Windowing of signal */
    accum_t acc1, acc2;
    if (mode)
    {
        ptr_sig = &x[0];
        ptr_base = state->b_mem_speech;
        for (i = 0; i < len; i += 4) chess_loop_range(2,HAMWINDLEN/4)
        {
            /* y[i] = mult_r(x[i], G722PLC_lpc_win_80[HAMWINDLEN-len+i]); / * for length < 80, uses the end of the
             * window * / */
            acc1 = fract_mult(*ptr_sig,   G722PLC_lpc_win_80[HAMWINDLEN - len + i]);/* for length < 80, uses the end of
                                                                                    * the window */
            ptr_sig = cyclic_add(ptr_sig,   1,   ptr_base, L_MEM_SPEECH);
            acc2 = fract_mult(*ptr_sig, G722PLC_lpc_win_80[HAMWINDLEN - len + i + 1]); /* for length < 80, uses the end
                                                                                        * of the window */
            ptr_sig = cyclic_add(ptr_sig,   1,   ptr_base, L_MEM_SPEECH);
            y[i]   = rnd_saturate(acc1);/* for length < 80, uses the end of the window */
            y[i + 1] = rnd_saturate(acc2); /* for length < 80, uses the end of the window */
            acc1 = fract_mult(*ptr_sig, G722PLC_lpc_win_80[HAMWINDLEN - len + i + 2]); /* for length < 80, uses the end
                                                                                        * of the window */
            ptr_sig = cyclic_add(ptr_sig,   1,   ptr_base, L_MEM_SPEECH);
            acc2 = fract_mult(*ptr_sig, G722PLC_lpc_win_80[HAMWINDLEN - len + i + 3]); /* for length < 80, uses the end
                                                                                        * of the window */
            ptr_sig = cyclic_add(ptr_sig,   1,   ptr_base, L_MEM_SPEECH);
            y[i + 2] = rnd_saturate(acc1); /* for length < 80, uses the end of the window */
            y[i + 3] = rnd_saturate(acc2); /* for length < 80, uses the end of the window */
        }
    }
    else
    {
        for (i = 0; i < len; i += 4) chess_loop_range(2,HAMWINDLEN/4)
        {
            /* y[i] = mult_r(x[i], G722PLC_lpc_win_80[HAMWINDLEN-len+i]); / * for length < 80, uses the end of the
             * window * / */
            acc1 = fract_mult(x[i],   G722PLC_lpc_win_80[HAMWINDLEN - len + i]);/* for length < 80, uses the end of the
                                                                                * window */
            acc2 = fract_mult(x[i + 1], G722PLC_lpc_win_80[HAMWINDLEN - len + i + 1]); /* for length < 80, uses the end
                                                                                        * of the window */
            y[i]   = rnd_saturate(acc1);/* for length < 80, uses the end of the window */
            y[i + 1] = rnd_saturate(acc2); /* for length < 80, uses the end of the window */
            acc1 = fract_mult(x[i + 2], G722PLC_lpc_win_80[HAMWINDLEN - len + i + 2]); /* for length < 80, uses the end
                                                                                        * of the window */
            acc2 = fract_mult(x[i + 3], G722PLC_lpc_win_80[HAMWINDLEN - len + i + 3]); /* for length < 80, uses the end
                                                                                        * of the window */
            y[i + 2] = rnd_saturate(acc1); /* for length < 80, uses the end of the window */
            y[i + 3] = rnd_saturate(acc2); /* for length < 80, uses the end of the window */
        }
    }

    /* Compute r[0] and test for overflow */

    do
    {
        Overflow = 0;

        /* sum = L_mac(1, y[0], y[0]); / *  1 for Avoid case of all zeros * / */
        acc = to_accum(1); /*  1 for Avoid case of all zeros */
        for (i = 0; i < len; i++) /* **TOdo** rewrite to take advantage of dual MAC?? */
        {
            /* sum = L_mac_ov(sum, y[i], y[i]); */
            acc += fract_mult(y[i], y[i]);
        }
        sum = rnd_saturate(acc);

        /* printf("%02X\n",extract_ext(acc)); */
        if ((extract_ext(acc) > 0) || (extract_high(acc) < 0))
        {
            /* if ((extract_ext(acc)>0)) { */
            Overflow = 1;
        }

        /* If overflow divide y[] by 4 */

        if (Overflow != 0)
        {
            for (i = 0; i < len; i++)
            {
                /* y[i] = shr(y[i], 2); */
                y[i] = y[i] >> 2;
            }
        }
    }
    while (Overflow != 0);

    /* Normalization of r[0] */

    /* norm = norm_l(sum); */
    normx = norm(acc);

    /* sum = L_shl(sum, norm); */
    sum = sum << normx;

    /* L_Extract(sum, &r_h[0], &r_l[0]); / * Put in DPF format (see oper_32b) * / */
    r[0] = sum;

    /* r[1] to r[m] */

    for (i = 1; i <= ord; i++)
    {
        /* sum = L_mult(y[0], y[i]); */
        acc = fract_mult(y[0], y[i]);

        /* tmp = sub(len, i); */
        tmp = len - i;
        for (j = 1; j < tmp; j++) /* **TODO** rewrite to take advantage of dual MAC?? */
        {
            /* sum = L_mac(sum, y[j], y[j + i]); */
            acc += fract_mult(y[j], y[j + i]);

            /* sum = rnd_saturate(acc); */
        }

        /* sum = L_shl(sum, norm); */
        sum = rnd_saturate(acc);
        sum = sum << normx;

        /* L_Extract(sum, &r_h[i], &r_l[i]); */
        r[i] = sum;
    }
    free_(y);
}

/*----------------------------------------------------------------------
 * G722PLC_pitch_ol(signal, length, maxco, overfl_shft)
 * open-loop pitch estimation
 *
 * signal      (i) : pointer to signal buffer (including signal memory)
 * length      (i) : length of signal memory
 * maxco       (o) : maximal correlation
 * overlf_shft (o) : number of shifts
 *
 *---------------------------------------------------------------------- */

static Word16 G722PLC_pitch_ol(G722PLC_STATE *restrict state, Word16 *maxco, Word16 *overfl_shft)
{
    Word16 i, j, il, k;
    Word16 ind, ind2;
    Word32 w_ds_siga[MAXPIT2_DS];
    static int /* chess_storage(DMB) */ w_ds_sigb[MAXPIT2_DS];    /* must be static as local variables are stored in stack
                                                             * (DMA)) */

    Word32 corx_f, ener1_f, ener2_f;
    Word32 temp_f;
    Word16 valid = 0;    /*not valid for the first lobe */
    Word16 start_ind, end_ind, beg_last_per, end_last_per;
    Word16 e1, e2, co, em, norm_e, ne1;
    Word32 ener1n, ener2n;
    Word16 x1;
    Word32 y1; /*_hi, y1_lo; */
    Word32 *ptr1;
    Word16 *signal;

    /* Word32 *nooffsigptr; */
    int    /* chess_storage(DMB) */ *nooffsigptr;
    Word32 L_temp;
    Word16 maxco_s8, stable;

    Word16 nooffsig[MAXPIT2 + FEC_L_FIR_FILTER_LTP_M1];
    Word32 ds_sig[MAXPIT2_DS];
    static int /* chess_storage(DMB) */ ai[4]; /* must be static as local variables are stored in stack (DMA)) */

    Word32 cor[3], rc[3]; /*cor_h[3], cor_l[3], */
    /*ai = (Word32x *)calloc_(4, sizeof(Word32x)); */

    /* signal = state->mem_speech + L_MEM_SPEECH - MAXPIT2; */
    signal = cyclic_add(state->mem_speech, L_MEM_SPEECH - MAXPIT2, state->b_mem_speech, L_MEM_SPEECH);

    /* intialize memory of DC-remove filter */

    /* for (i = 0; i < FEC_L_FIR_FILTER_LTP_M1; i++)
     * {
     * nooffsig[i] = (Word16) 0;
     * } */
    nooffsigptr = (int    /* chess_storage(DMB) */ *)(nooffsig + FEC_L_FIR_FILTER_LTP_M1);

    /* DC-remove filter */
    x1 = y1 = 0; /*_hi = y1_lo = 0; */

    for (i = 0; i < MAXPIT2; i++)
    {
        /* nooffsigptr[i] = G722PLC_hp50(&x1, &y1, signal[i]); */
        nooffsigptr[i] = G722PLC_hp50(&x1, &y1, *signal);
        signal = cyclic_add(signal, 1, state->b_mem_speech, L_MEM_SPEECH);
    }

    /* downsample (filter and decimate) signal */
    ptr1 = ds_sig;
    accum_t dec_acc;
    accum_t dec_acc1;
    for (i = FACT_M1; i < MAXPIT2; i += FACT)
    {
/*		temp_f = L_mult0(nooffsigptr[i], G722PLC_fir_lp[0]); */
        dec_acc = to_accum(1);
        for (k = 0; k < FEC_L_FIR_FILTER_LTP - 1; k++)
        {
/*			temp_f = L_mac0(temp_f, nooffsigptr[sub(i, k)], G722PLC_fir_lp[k]); */
            dec_acc += fract_mult(nooffsigptr[i - k],        G722PLC_fir_lp[k]);
        }
        dec_acc += fract_mult(nooffsigptr[i - FEC_L_FIR_FILTER_LTP + 1],   G722PLC_fir_lp[FEC_L_FIR_FILTER_LTP - 1]);

/*     *ptr1++ = round(temp_f); */
        *ptr1++ = rnd_saturate(dec_acc);
    }

    /* G722PLC_autocorr(ds_sig, cor_h, cor_l, 2, MAXPIT2_DS);
     * G722PLC_lag_window(cor_h, cor_l, 2);	/ * Lag windowing		* /
     * G722PLC_levinson(cor_h, cor_l, rc, &stable, 2, ai); */

    G722PLC_autocorr(ds_sig, cor, 2, MAXPIT2_DS, 0, NULL);
    G722PLC_lag_window(cor, 2); /* Lag windowing		*/
    G722PLC_levinson(cor, rc, &stable, 2, ai);

    /* ai[1] = round(L_mult(ai[1],GAMMA));
     * ai[2] = round(L_mult(ai[2],GAMMA2)); */
    ai[1] = rnd_saturate(fract_mult(ai[1], GAMMA));
    ai[2] = rnd_saturate(fract_mult(ai[2], GAMMA2));

    /* filter */
    accum_t filt_acc1;
    accum_t filt_acc2;

    /* w_ds_sig = (Word32 *)calloc_(MAXPIT2_DS, sizeof(Word32)); */
    w_ds_siga[0] = ds_sig[0];
    w_ds_sigb[0] = ds_sig[0];

    /* L_temp = L_mult(ai[1], ds_sig[0]);
     * w_ds_sig[1] = add(ds_sig[1], round(L_shl(L_temp,3))); */
    filt_acc1 = fract_mult(ai[1], ds_sig[0]);

    /* w_ds_sig[1] = add(ds_sig[1], rnd_saturate(filt_acc<<3)>>16); */
    w_ds_siga[1] = rnd_saturate(to_accum(ds_sig[1]) + (filt_acc1 << 3));
    w_ds_sigb[1] = w_ds_siga[1];

    for (i = 2; i < MAXPIT2_DS; i += 2)
    {
        /* L_temp = L_mult(ai[1], ds_sig[i - 1]);
         * L_temp = L_mac(L_temp, ai[2], ds_sig[i - 2]);
         * w_ds_sig[i] = add(ds_sig[i], round(L_shl(L_temp,3))); */
        filt_acc1  = fract_mult(ai[1], ds_sig[i - 1]);
        filt_acc1 += fract_mult(ai[2], ds_sig[i - 2]);

        /* w_ds_sig[i] = add(ds_sig[i], rnd_saturate(filt_acc<<3)>>16);
         * L_temp = rnd_saturate(filt_acc); */
        w_ds_siga[i] = rnd_saturate(to_accum(ds_sig[i]) + (filt_acc1 << 3));
        w_ds_sigb[i] = w_ds_siga[i];

        filt_acc2  = fract_mult(ai[1], ds_sig[i]);
        filt_acc2 += fract_mult(ai[2], ds_sig[i - 1]);
        w_ds_siga[i + 1] = rnd_saturate(to_accum(ds_sig[i + 1]) + (filt_acc2 << 3));
        w_ds_sigb[i + 1] = w_ds_siga[i + 1];
    }

    ind = MAXPIT_S2_DS;    /*default value*/
    ind2 = 0;

    /*Test overflow on w_ds_sig*/
    *overfl_shft = 0;

    /* compute energy of signal in range [len/fac-1,(len-MAX_PIT)/fac-1] */
    ener1_f = 1;
    accum_t ener1_f_acc = to_accum(ener1_f);

    for (j = MAXPIT2_DSM1; j >= MAXPIT_DSP1; j--)
    {
        /* ener1_f = L_mac0(ener1_f, w_ds_sig[j], w_ds_sig[j]); */
        ener1_f_acc += fract_mult(w_ds_siga[j], w_ds_sigb[j]);
    }

    /* ener1_f=rnd_saturate(ener1_f_acc); */

    /* compute exponent */
    ne1 = norm(ener1_f_acc);

    /* compute maximal correlation (maxco) and pitch lag (ind) */
    *maxco = 0;
    accum_t corx_f_acc;
    accum_t ener2_f_acc;

    /* ener2_f = L_msu0(ener1_f, w_ds_sig[MAXPIT2_DSM1], w_ds_sig[MAXPIT2_DSM1]); / *update, part 1* / */
    ener2_f_acc = ener1_f_acc - fract_mult(w_ds_siga[MAXPIT2_DSM1], w_ds_sigb[MAXPIT2_DSM1]);    /*update, part 1*/
    for (i = 1; i < MAXPIT_DS; i++)    /* < to avoid out of range later*/
    {
        ind2 = add(ind2, 1);
        corx_f_acc = to_accum(0);

        for (j = MAXPIT2_DSM1; j >= MAXPIT_DSP1; j--)
        {
            /* corx_f = L_mac0(corx_f, w_ds_sig[j], w_ds_sig[j-i]); */
            corx_f_acc += fract_mult(w_ds_siga[j], w_ds_sigb[j - i]);

            /* corx_f = rnd_saturate(corx_f_acc); */
        }

        /* ON Semi update =============== */

        ener2_f_acc += fract_mult(w_ds_siga[MAXPIT_DSP1 - i], w_ds_sigb[MAXPIT_DSP1 - i]);    /*update, part 2*/
        norm_e = min(ne1, norm(ener2_f_acc));

        /* corx_f = rnd_saturate(ener2_f_acc);
         * temp_f = rnd_saturate(ener1_f_acc); */
        ener2_f_acc -= fract_mult(w_ds_siga[MAXPIT2_DSM1 - i], w_ds_sigb[MAXPIT2_DSM1 - i]);         /*update, part 1*/
        e1 = rnd_saturate(ener1_f_acc << norm_e);
        e2 = rnd_saturate(ener2_f_acc << norm_e);
        corx_f_acc =  corx_f_acc << (norm_e);
        co = rnd_saturate(corx_f_acc) >> 16;
        em = max(e1, e2) >> 16;

        /* ===============================
         * original code
         * ener2_f = L_mac0(ener2_f, w_ds_sig[MAXPIT_DSP1-i], w_ds_sig[MAXPIT_DSP1-i]); / *update, part 2* /
         * norm_e = s_min(ne1, norm_l(ener2_f));
         * ener2_f = L_msu0(ener2_f, w_ds_sig[MAXPIT2_DSM1-i], w_ds_sig[MAXPIT2_DSM1-i]);		 / *update, part 1* /
         * ener1n = L_shl(ener1_f, norm_e);
         * ener2n = L_shl(ener2_f, norm_e);
         * corx_f = L_shl(corx_f, norm_e);
         * e1 = round(ener1n);
         * e2 = round(ener2n);
         * co = round(corx_f);
         * em = s_max(e1, e2); */

        if (sub(co, em) > 0)
        {
            em = co;

/*          move16(); */
        }
        if (co > 0)
        {
            /* co = div_s(co, em); */
            co = (co << 15) / em;
        }

        if (co < 0)
        {
            valid = 1;
        }

        /* if (sub(valid,1) == 0) */
        if (valid == 1)
        {
            /* if ((sub(ind2, ind) == 0) || (sub(ind2, shl(ind,1)) == 0)) */
            if ((ind2 == ind) || (ind2 == (ind << 1)))
            {
                /* if(sub(*maxco, 27850) > 0) / * 0.85 : high correlation, small chance that double pitch is OK* / */
                if (*maxco > 27850) /* 0.85 : high correlation, small chance that double pitch is OK*/
                {
                    *maxco = 32767;
                }

                /* maxco_s8 = shr(*maxco, 3); */
                maxco_s8 = *maxco >> 3;

                /* if(sub(*maxco, 29126) < 0)/ *to avoid overflow* / */
                if (*maxco < 29126) /*to avoid overflow*/
                {
                    /* *maxco = add(*maxco, maxco_s8); */
                    *maxco = *maxco + maxco_s8;
                }
            }

            /* if ((sub(co, *maxco) > 0) && (sub(i, MINPIT_DS) >= 0)) */
            if ((co > *maxco) && (i >= MINPIT_DS))
            {
                *maxco = co;
                ind = i;
                ind2 = 1;
            }
        }
    }

    /* free_(w_ds_sig); */

    /* convert pitch to non decimated domain */

    /* il = shl(ind, FACTLOG2); */
    il = ind << FACTLOG2;
    ind = il;

    /* shift DC-removed signal to avoid overflow in correlation */

    /* if(L_sub(ener1_f, 0x01000000) > 0) / * maxcor will be computed on 4* points in non weighted domain --> overflow
     * risq* /
     * if(ener1_f > 0x01000000<<1) / * maxcor will be computed on 4* points in non weighted domain --> overflow risq* /
     * {
     * // *overfl_shft = add(*overfl_shft,1);
     * *overfl_shft++;
     * } */

    if (*overfl_shft > 0)
    {
        for (i = 1; i < MAXPIT2; i++)
        {
            /* nooffsigptr[i] = shr(nooffsigptr[i], *overfl_shft); */
            nooffsigptr[i] = nooffsigptr[i] >> *overfl_shft;
        }
    }

    /* refine pitch in non-decimated (8 kHz) domain by step of 1
     * -> maximize correlation around estimated pitch lag (ind) */

    /* start_ind = sub(il, 2);
     * start_ind = s_max(start_ind, MINPIT);
     * end_ind = add(il, 2);
     * beg_last_per = sub(MAXPIT2, il);
     * end_last_per = sub(MAXPIT2, 1); */
    start_ind = il - 2;
    start_ind = max(start_ind, MINPIT);
    end_ind = il + 2;
    beg_last_per = MAXPIT2 - il;
    end_last_per = MAXPIT2 - 1;

/*  j = sub(end_last_per, start_ind); */
    /* ener1_f = L_mac0(1, nooffsigptr[end_last_per], nooffsigptr[end_last_per]); / *to avoid division by 0* /
     * ener2_f = L_mac0(1, nooffsigptr[j], nooffsigptr[j]); / *to avoid division by 0* / */
    ener1_f_acc = to_accum(1); /*to avoid division by 0*/
    ener2_f_acc = to_accum(1); /*to avoid division by 0*/
    for (j = end_last_per; j >= beg_last_per; j--)
    {
        /* ener1_f	= L_mac0(ener1_f, nooffsigptr[j], nooffsigptr[j]);
         * ener2_f	= L_mac0(ener2_f, nooffsigptr[j-start_ind], nooffsigptr[j-start_ind]); */
        ener1_f_acc    += fract_mult(nooffsigptr[j], nooffsigptr[j]);
        ener2_f_acc    += fract_mult(nooffsigptr[j - start_ind], nooffsigptr[j - start_ind]);

        /* ener2_f = rnd_saturate(ener2_f_acc>>1); */
    }

    /* j = sub(beg_last_per, start_ind); */
    j = beg_last_per - start_ind;

    /* ener2_f = L_msu0(ener2_f, nooffsigptr[j], nooffsigptr[j]); / *to compansate first update part 2* / */
    ener2_f_acc -= fract_mult(nooffsigptr[j], nooffsigptr[j]);    /*to compansate first update part 2*/
    /* compute exponent */

    /* ener1_f = rnd_saturate(ener1_f_acc); */
    ne1 = norm(ener1_f_acc);

    /* compute maximal correlation (maxco) and pitch lag (ind) */
    *maxco = 0;

    for (i = start_ind; i <= end_ind; i++)
    {
        corx_f = 0;
        corx_f_acc = to_accum(0);

        /* ener2_f = L_mac0(ener2_f, nooffsigptr[beg_last_per-i], nooffsigptr[beg_last_per-i]); / *update, part 2* / */
        ener2_f_acc += fract_mult(nooffsigptr[beg_last_per - i], nooffsigptr[beg_last_per - i]);    /*update, part 2*/

        /* ener2_f = rnd_saturate(ener2_f_acc>>1); */
        for (j = end_last_per; j >= beg_last_per; j--)
        {
            /* corx_f	= L_mac0(corx_f, nooffsigptr[j], nooffsigptr[j-i]); */
            corx_f_acc  += fract_mult(nooffsigptr[j], nooffsigptr[j - i]);
        }

        /* ON Semi update =============== */
        norm_e = min(ne1, norm(ener2_f_acc));
        e1 = rnd_saturate(ener1_f_acc << norm_e);
        e2 = rnd_saturate(ener2_f_acc << norm_e);
        corx_f_acc =  corx_f_acc << (norm_e);
        co = rnd_saturate(corx_f_acc) >> 16;
        em = max(e1, e2) >> 16;

        /* ===============================
         * original code
         * norm_e = s_min(ne1, norm_l(ener2_f));
         * ener1n = L_shl(ener1_f, norm_e);
         * ener2n = L_shl(ener2_f, norm_e);
         * corx_f = L_shl(corx_f, norm_e);
         * e1 = round(ener1n);
         * e2 = round(ener2n);
         * co = round(corx_f);
         * em = s_max(e1, e2);
         * =============================== */

        /* if(sub(co, em) > 0) */
        if (co > em)
        {
            em = co;
        }
        if (co > 0)
        {
            /* co = div_s(co, em); */
            co = (co << 15) / em;
        }

        /* if(sub(co, *maxco) > 0) */
        if (co > *maxco)
        {
            ind = i;
        }

        /* *maxco = s_max(co, *maxco); */
        *maxco = max(co, *maxco);

        /* ener2_f = L_msu0(ener2_f, nooffsigptr[end_last_per-i], nooffsigptr[end_last_per-i]);		 / *update, part 1* / */
        ener2_f_acc -= fract_mult(nooffsigptr[end_last_per - i], nooffsigptr[end_last_per - i]); /*update, part 1*/
    }

    /* ener2_f = rnd_saturate(ener2_f_acc>>1); */

    /* if (sub(*maxco, 8192) < 0) */
    if (*maxco < 8192)
    {
        /* if(sub(ind, 32) < 0) */
        if (ind < 32)
        {
            /* ind = shl(ind,1); / *2 times pitch for very small pitch, at least 2 times MINPIT * / */
            ind = ind << 1;    /*2 times pitch for very small pitch, at least 2 times MINPIT */
        }
    }

    return ind;
}

/*----------------------------------------------------------------------
 * G722PLC_classif_modif(maxco, overfl_shft, decoder)
 * signal classification and conditional residual modification
 *
 * maxco       (i) : maximal correlation
 * nbl         (i) : lower-band G722 scale factor
 * nbh         (i) : higher-band G722 scale factor
 * state       (i) : pointer to PLC state
 *---------------------------------------------------------------------- */

static Word16 G722PLC_classif_modif(Word16 maxco, Word16 nbl, Word16 nbh, G722PLC_STATE *restrict state)
{
    Word16 *mem_speech;       /* pointer to speech buffer */
    Word16 l_mem_speech;      /* length of speech buffer */
    Word16 *mem_exc;          /* pointer to excitation buffer */
    Word16 l_exc;             /* length of excitation buffer */
    Word16 clas, Temp, tmp1, tmp2, tmp3, tmp4, i, maxres, absres, zcr, clas2;
    Word16 *pt1, *pt2;

    /************************************************************************
    * select preliminary class => clas = UNVOICED, WEAKLY_VOICED or VOICED *
    * by default clas=WEAKLY_VOICED                                                         *
    * classification criterio:                                             *
    * -normalized correlation from open-loop pitch                         *
    * -ratio of lower/higher band energy (G722 scale factors)              *
    * -zero crossing rate                                                  *
    ************************************************************************/
    mem_speech   = state->mem_speech;
    l_mem_speech = L_MEM_SPEECH;
    mem_exc      = state->mem_exc;
    l_exc        = L_EXC;

    /* compute zero-crossing rate in last 10 ms */

    /* pt1 = &mem_speech[sub(l_mem_speech, 80)];
     * pt2 = pt1-1; */
    pt1 = cyclic_add(mem_speech, l_mem_speech - 80, state->b_mem_speech, l_mem_speech);
    pt2 = cyclic_add(pt1, -1, state->b_mem_speech, l_mem_speech);
    zcr = 0;

    for (i = 0; i < 80; i++)
    {
        Temp = 0;

        if (*pt1 <= 0)
        {
            Temp = 1;
        }
        if (*pt2 > 0)
        {
            /* Temp = add(Temp,1); */
            Temp++;
        }

        /* zcr = add(zcr, shr(Temp,1)); */
        zcr = zcr + (Temp >> 1);

        /* pt1++;
         * pt2++; */
        pt1 = cyclic_add(pt1, 1, state->b_mem_speech, l_mem_speech);
        pt2 = cyclic_add(pt2, 1, state->b_mem_speech, l_mem_speech);
    }

    /* set default clas */
    clas = G722PLC_WEAKLY_VOICED;

    /* detect voiced clas (corr > 3/4 ener1 && corr > 3/4 ener2) */

    /* if(sub(maxco, 22936) > 0) / * 22936 in Q15 = 0.7 * / */
    if (maxco > 22936) /* 22936 in Q15 = 0.7 */
    {
        clas = G722PLC_VOICED;
    }

    /* change class to unvoiced if higher band has lots of energy
     * (risk of "dzing" if clas is "very voiced") */

    /* if (sub(nbh, nbl) > 0) */
    if (nbh > nbl)
    {
        clas2 = clas;
        clas = G722PLC_VUV_TRANSITION;

        /* if(sub(clas2, G722PLC_VOICED) == 0) */
        if (sub(clas2, G722PLC_VOICED) == 0)
        {
            clas = G722PLC_WEAKLY_VOICED;
        }
    }

    /* change class to unvoiced if zcr is high */

    /* if (sub(zcr,20)>=0) */
    if (zcr >= 20)
    {
        clas = G722PLC_UNVOICED;

        /* change pitch if unvoiced class (to avoid short pitch lags) */

        /* if(sub(state->t0, 32) < 0) */
        if (state->t0 < 32)
        {
            /* state->t0 = shl(state->t0,1); / *2 times pitch for very small pitch, at least 2 times MINPIT * / */
            state->t0 = state->t0 << 1; /*2 times pitch for very small pitch, at least 2 times MINPIT */
        }
    }

    /**************************************************************************
    * detect transient => clas = TRANSIENT	                                  *
    * + modify residual to limit amplitude for LTP                           *
    * (this is performed only if current class is not VOICED to avoid        *
    *  perturbation of the residual for LTP)                                 *
    **************************************************************************/

    /* detect transient and limit amplitude of residual */
    Temp = 0;

    /* if (sub(clas,4) > 0)/ *G722PLC_WEAKLY_VOICED(5) or G722PLC_VUV_TRANSITION(7)* / */
    if (clas > 4) /*G722PLC_WEAKLY_VOICED(5) or G722PLC_VUV_TRANSITION(7)*/
    {
        /* tmp1 = sub(l_exc, state->t0); / * tmp1 = start index of last "pitch cycle" * /
         * tmp2 = sub(tmp1, state->t0);  / * tmp2 = start index of last but one "pitch cycle" * / */
        tmp1 = l_exc - state->t0; /* tmp1 = start index of last "pitch cycle" */
        tmp2 = tmp1  - state->t0;/* tmp2 = start index of last but one "pitch cycle" */
        for (i = 0; i < state->t0; i++)
        {
            /* tmp3 = add(tmp2, i); */
            tmp3 = tmp2 + i;

            /* maxres = s_max(s_max(abs_s(mem_exc[tmp3-2]), abs_s(mem_exc[tmp3-1])),
             * s_max(abs_s(mem_exc[tmp3]),   abs_s(mem_exc[tmp3+1])));
             * maxres = s_max(abs_s(mem_exc[tmp3+2]), maxres);
             * absres = abs_s(mem_exc[tmp1 + i]); */
            maxres = max(max(abs(mem_exc[tmp3 - 2]), abs(mem_exc[tmp3 - 1])),
                         max(abs(mem_exc[tmp3]),   abs(mem_exc[tmp3 + 1])));
            maxres = max(abs(mem_exc[tmp3 + 2]), maxres);
            absres = abs(mem_exc[tmp1 + i]);

            /* if magnitude in last cycle > magnitude in last but one cycle */

            /* if (sub(absres, maxres) > 0) */
            if (absres > maxres)
            {
                /* detect transient (ratio = 1/8) */

                /* tmp4 = shr(absres,3); */
                tmp4 = absres >> 3;

                /* if(L_sub(tmp4, maxres) >= 0) */
                if (tmp4 >= maxres)
                {
                    /* Temp = add(Temp, 1); */
                    Temp++;
                }

                /* limit value (even if a transient is not detected...) */
                if (mem_exc[tmp1 + i] < 0)
                {
                    /* mem_exc[tmp1 + i] = negate(maxres); */
                    mem_exc[tmp1 + i] = -(maxres);
                }
                if (mem_exc[tmp1 + i] >= 0)
                {
                    mem_exc[tmp1 + i] = maxres;
                }
            }
        }
    }
    if (Temp > 0)
    {
        clas = G722PLC_TRANSIENT;
    }

    return clas;
}

/*----------------------------------------------------------------------
 * G722PLC_syn_filt_circ(m, a, x, y, n state)
 * LPC synthesis filter
 *
 * m (i) : LPC order
 * a (i) : LPC coefficients
 * x (i) : input buffer
 * y (o) : output buffer
 * n (i) : number of samples
 *---------------------------------------------------------------------- */

static inline void G722PLC_syn_filt_circ(Word16 m, Word32x *a, Word16 *x, Word16 *y, G722PLC_STATE *state)
{
    Word32 L_temp;
    volatile Word16 j;
    accum_t acc;
    Word16  *ptr;

    /* L_temp = L_mult(a[0], *x);  / * Q28= Q12 * Q15 * 2 * / */
    acc = fract_mult(a[0], *x);
    ptr = cyclic_add(y, -1, state->b_mem_speech, L_MEM_SPEECH);
    for (j = 1; j <= m; j++)
    {
        /* L_temp = L_msu(L_temp, a[j], y[-j]);  / * Q28= Q12 * Q15 * 2 * / */
        acc -= fract_mult(a[j], *ptr); /* Q28= Q12 * Q15 * 2 */
        ptr = cyclic_add(ptr, -1, state->b_mem_speech, L_MEM_SPEECH);
    }

    /* *y = round(L_shl(L_temp, 3)); */
    *y = rnd_saturate(acc << 3);
}

/*----------------------------------------------------------------------
 * G722PLC_syn_filt(m, a, x, y, n n)
 * LPC synthesis filter
 *
 * m (i) : LPC order
 * a (i) : LPC coefficients
 * x (i) : input buffer
 * y (o) : output buffer
 * n (i) : number of samples
 *---------------------------------------------------------------------- */

static inline void G722PLC_syn_filt(Word16 m, Word32x *a, Word16 *x, Word16 *y)
{
    Word32 L_temp;
    volatile Word16 j;
    accum_t acc;

    /* L_temp = L_mult(a[0], *x);  / * Q28= Q12 * Q15 * 2 * / */
    acc = fract_mult(a[0], *x); /* Q28= Q12 * Q15 * 2 */
    for (j = 1; j <= m; j++)
    {
        /* L_temp = L_msu(L_temp, a[j], y[-j]);  / * Q28= Q12 * Q15 * 2 * / */
        acc -= fract_mult(a[j], y[-j]); /* Q28= Q12 * Q15 * 2 */

        /* L_temp = rnd_saturate(acc); */
    }

    /* *y = round(L_shl(L_temp, 3)); */
    *y = rnd_saturate(acc << 3);
}

/*----------------------------------------------------------------------
 * G722PLC_ltp_pred_1s(cur_exc, t0, jitter)
 * one-step LTP prediction and jitter update
 *
 * exc     (i)   : excitation buffer (exc[...-1] correspond to past)
 * t0      (i)   : pitch lag
 * jitter  (i/o) : pitch lag jitter
 *---------------------------------------------------------------------- */
#if 0    /* no longer used */
static inline Word16 G722PLC_ltp_pred_1s(Word16 *exc, Word16 t0, Word16 *jitter)
{
    Word16 i;

    /* i = sub(*jitter, t0); */
    i = *jitter - t0;

    /* update jitter for next sample */

    /* *jitter = negate(*jitter); */
    *jitter = rnd_saturate(to_accum(-*jitter));

    /* prediction =  exc[-t0+jitter] */
    return exc[i];
}

#endif /* if 0 */

/*----------------------------------------------------------------------
 * G722PLC_ltp_syn(state, cur_exc, cur_syn, n, jitter)
 * LTP prediction followed by LPC synthesis filter
 *
 * state    (i/o) : PLC state variables
 * cur_exc  (i)   : pointer to current excitation sample (cur_exc[...-1] correspond to past)
 * cur_syn  (i/o) : pointer to current synthesis sample
 * n     (i)      : number of samples
 * jitter  (i/o)  : pitch lag jitter
 *---------------------------------------------------------------------- */

static void G722PLC_ltp_syn(G722PLC_STATE *restrict state, Word16 *cur_exc, Word16 *cur_syn, Word16 n, Word16 *jitter)
{
    Word16 i;
    Word16 k;
    Word16 loc_jitter;
    Word16 p_idx;
    Word16 n_idx;
    Word16 *cur_exc1;
    Word16 *cur_exc2;

    loc_jitter = *jitter;
    p_idx = loc_jitter - state->t0;
    n_idx = -loc_jitter - state->t0;

    for (i = 0; i < n / 2; i++)
        chess_loop_range(2, CROSSFADELEN)
        {
            /* LTP prediction using exc[...-1] */

            /* *cur_exc = G722PLC_ltp_pred_1s(cur_exc, state->t0, jitter); */
            *cur_exc = cur_exc[p_idx];

            /* LPC synthesis filter (generate one sample) */
            G722PLC_syn_filt(ORD_LPC, state->a, cur_exc, cur_syn);

            cur_exc++;
            cur_syn++;

            *cur_exc = cur_exc[n_idx];

            /* LPC synthesis filter (generate one sample) */
            G722PLC_syn_filt(ORD_LPC, state->a, cur_exc, cur_syn);

            cur_exc++;
            cur_syn++;
        }
        *jitter = loc_jitter;
}

/*----------------------------------------------------------------------
 * G722PLC_syn(state, syn, n)
 * extrapolate missing lower-band signal (PLC)
 *
 * state (i/o) : PLC state variables
 * syn   (o)   : synthesis
 * n     (i)   : number of samples
 *---------------------------------------------------------------------- */

static void G722PLC_syn(G722PLC_STATE *state, Word32 *syn, Word16 n, Word16 cyclic_mode)
{
	Word16	 *buffer_syn; /* synthesis buffer */
	Word16	 *buffer_exc; /* excitation buffer */
    Word16   *cur_syn;    /* pointer to current sample of synthesis */
    Word16   *cur_exc;    /* pointer to current sample of excition */
    Word16   *exc;        /* pointer to beginning of excitation in current frame */
    Word16 temp;
    Word16 jitter, dim;
    Word16 p_idx;
    Word16 n_idx;

    /* dim = add(n, add(state->t0, T0SAVEPLUS)); */
    dim = n + state->t0 + T0SAVEPLUS;

    /* allocate temporary buffers and set pointers */

    buffer_exc = (Word16 *)malloc_(dim*sizeof(Word16));
    buffer_syn = (Word16 *)malloc_(2*ORD_LPC*sizeof(Word16)); // minimal allocations of scratch RAM

    /* cur_exc = &buffer_exc[add(state->t0, T0SAVEPLUS)]; */
    cur_exc = &buffer_exc[state->t0 + T0SAVEPLUS];
    cur_syn = &buffer_syn[ORD_LPC];

    exc = cur_exc;

    /* copy memory
     * - past samples of synthesis (LPC order)            -> buffer_syn[0]
     * - last "pitch cycle" of excitation (t0+t0SavePlus) -> buffer_exc[0]
     */
    G722PLC_copy(state->mem_syn, buffer_syn, ORD_LPC); /*  */

    /* temp = add(state->t0, T0SAVEPLUS); */
    temp = state->t0 + T0SAVEPLUS;

    /* G722PLC_copy(state->mem_exc + sub(L_EXC, temp), buffer_exc, temp); */
    G722PLC_copy(state->mem_exc + L_EXC - temp, buffer_exc, temp);

    /***************************************************
    * set pitch jitter according to clas information *
    ***************************************************/

    jitter = s_and(state->clas, 1);

    /* state->t0 = s_or(state->t0, jitter);    / * change even delay as jitter is more efficient for odd delays * / */
    state->t0 = state->t0 | jitter;   /* change even delay as jitter is more efficient for odd delays */

    /*****************************************************
    * generate signal by LTP prediction + LPC synthesis *
    *****************************************************/

    /* temp = sub(n, ORD_LPC); */
    temp = n - ORD_LPC;

    /* first samples [0...ord-1] */
    G722PLC_ltp_syn(state, cur_exc, cur_syn, ORD_LPC, &jitter);
    Word16 *ptr1, *ptr2;
    ptr1 = syn;
    ptr2 = cur_syn;
    if (cyclic_mode)
    {
        for (int i = 0; i < ORD_LPC; i++)
        {
            *ptr1 = *ptr2;
            ptr2++;

            /* ptr1 = cyclic_add(ptr1, 1, state->b_mem_speech, L_MEM_SPEECH); */
            ptr1 = cyclic_add(ptr1, 1, state->b_mem_speech, L_MEM_SPEECH);
        }

        /* remaining samples [ord...n-1] */

        /* ptr1 = &syn[ORD_LPC]; */
        ptr1 = cyclic_add(syn, ORD_LPC, state->b_mem_speech, L_MEM_SPEECH);
        ptr2 = &cur_exc[ORD_LPC];
        p_idx = jitter - state->t0;
        n_idx = -jitter - state->t0;

        /* G722PLC_ltp_syn(state, &cur_exc[ORD_LPC], &syn[ORD_LPC], temp, &jitter); */
        for (int i = 0; i < temp / 2; i++)
            chess_loop_range(2, CROSSFADELEN)
            {
                /* LTP prediction using exc[...-1] */

                /* *ptr2 = G722PLC_ltp_pred_1s(ptr2, state->t0, &jitter); */
                *ptr2 = ptr2[p_idx];

                /* LPC synthesis filter (generate one sample) */
                G722PLC_syn_filt_circ(ORD_LPC, state->a, ptr2++, ptr1, state);

                ptr1 = cyclic_add(ptr1, 1, state->b_mem_speech, L_MEM_SPEECH);

                /* LTP prediction using exc[...-1] */

                /* *ptr2 = G722PLC_ltp_pred_1s(ptr2, state->t0, &jitter); */
                *ptr2 = ptr2[n_idx];

                /* LPC synthesis filter (generate one sample) */
                G722PLC_syn_filt_circ(ORD_LPC, state->a, ptr2++, ptr1, state);

                ptr1 = cyclic_add(ptr1, 1, state->b_mem_speech, L_MEM_SPEECH);
            }

            /* update memory:
             * - synthesis for next frame (last LPC-order samples)
             * - excitation */

            /* G722PLC_copy(&syn[temp], state->mem_syn, ORD_LPC); */
            ptr1 = cyclic_add(syn, temp, state->b_mem_speech, L_MEM_SPEECH);
        ptr2 = state->mem_syn;
        for (int i = 0; i < ORD_LPC; i++)
        {
            *ptr2 = *ptr1;
            ptr2++;
            ptr1 = cyclic_add(ptr1, 1, state->b_mem_speech, L_MEM_SPEECH);
        }
    }
    else
    {
        G722PLC_copy(cur_syn, syn, ORD_LPC);

        /* remaining samples [ord...n-1] */
        G722PLC_ltp_syn(state, &cur_exc[ORD_LPC], &syn[ORD_LPC], temp, &jitter);

        /* update memory:
         * - synthesis for next frame (last LPC-order samples)
         * - excitation */
        G722PLC_copy(&syn[temp], state->mem_syn, ORD_LPC);
    }

    G722PLC_update_mem_exc(state, exc, n);

    /* free allocated memory */

	free_(buffer_syn);
	free_(buffer_exc);
}

/*-------------------------------------------------------------------------*
 * Function G722PLC_lpc																	 *
 *--------------------------------------------------------------------------*/
static void G722PLC_lpc(G722PLC_STATE *restrict state)
{
    Word16 tmp;

    /* Word16		cor_h[ORD_LPC + 1];
     * Word16		cor_l[ORD_LPC + 1]; */
    Word16 cor[ORD_LPC + 1];
    Word16 rc[ORD_LPC + 1];
    Word16      *ptr;

/* printf("G722PLC_lpc\n"); */
    /* G722PLC_autocorr(&state->mem_speech[L_MEM_SPEECH - HAMWINDLEN], cor_h, cor_l, ORD_LPC, HAMWINDLEN);
     * G722PLC_lag_window(cor_h, cor_l, ORD_LPC);	/ * Lag windowing		* /
     * G722PLC_levinson(cor_h, cor_l, rc, &tmp, ORD_LPC, state->a); */
    ptr = cyclic_add(state->mem_speech, L_MEM_SPEECH - HAMWINDLEN, state->b_mem_speech, L_MEM_SPEECH);
    G722PLC_autocorr(&ptr[0], cor,  ORD_LPC, HAMWINDLEN, 1, state);
    G722PLC_lag_window(cor, ORD_LPC);    /* Lag windowing		*/
    G722PLC_levinson(cor, rc, &tmp, ORD_LPC, state->a);

#ifdef USE_FEC_PRINTF
    if (stable != 0)
    {
        fprintf(stderr, "WARNING: LPC filter unstable !!!\n");

        /*exit(0); */
    }
#endif /* ifdef USE_FEC_PRINTF */
}

/*-------------------------------------------------------------------------*
 * Function G722PLC_residu                                                              *
 *--------------------------------------------------------------------------*/
static void G722PLC_residu(G722PLC_STATE *restrict state)
{
    Word32 L_temp;
    Word16   *ptr_sig, *ptr_res, *ptr_init, *ptr_base;
    volatile Word16 i, j;
    accum_t acc1, acc2;

    ptr_res = state->mem_exc + sub(L_EXC, MAXPIT2P1);

    /* ptr_sig =
     * &state->mem_speech[L_MEM_SPEECH - MAXPIT2P1]; */
    ptr_sig = cyclic_add(state->mem_speech, L_MEM_SPEECH - MAXPIT2P1, state->b_mem_speech, L_MEM_SPEECH);
    ptr_base = &state->b_mem_speech[0];
    ptr_init = ptr_sig;

    for (i = 0; i < MAXPIT2P1 - 1; i += 2)
    {
        /* L_temp = L_mult(ptr_sig[i], state->a[0]);
         * acc = fract_mult(ptr_sig[i], state->a[0]); */
        acc1 = to_accum(0);
        acc2 = to_accum(0);
        for (j = 0; j < ORD_LPC - 1; j += 2)
        {
            /* L_temp = L_mac(L_temp, state->a[j], ptr_sig[i - j]); */
            acc2 += fract_mult(state->a[j + 1], *ptr_sig);
            acc1 += fract_mult(state->a[j],   *ptr_sig);
            ptr_sig = cyclic_add(ptr_sig,   1,   ptr_base, L_MEM_SPEECH);
            acc2 += fract_mult(state->a[j],   *ptr_sig);
            ptr_sig = cyclic_add(ptr_sig, -2,   ptr_base, L_MEM_SPEECH);
            acc1 += fract_mult(state->a[j + 1], *ptr_sig);
            ptr_sig = cyclic_add(ptr_sig, -1,  ptr_base, L_MEM_SPEECH);
        }
        acc1 += fract_mult(state->a[8],   *ptr_sig);
        ptr_sig = cyclic_add(ptr_sig,   1,   ptr_base, L_MEM_SPEECH);
        acc2 += fract_mult(state->a[8],   *ptr_sig);

        ptr_sig = cyclic_add(ptr_init, i + 2, ptr_base, L_MEM_SPEECH);

        /* L_temp = L_shl(L_temp, 3);	/ * Q28 -> Q31 * / */
        ptr_res[i]     = rnd_saturate(acc1 << 3);    /*Q31 -> Q15 */
        ptr_res[i + 1] = rnd_saturate(acc2 << 3);    /*Q31 -> Q15 */
    }
    acc1 = to_accum(0);
    for (j = 0; j < ORD_LPC - 1; j += 2)
    {
        acc1 += fract_mult(state->a[j],   *ptr_sig);
        ptr_sig = cyclic_add(ptr_sig,  -1,   ptr_base, L_MEM_SPEECH);
        acc1 += fract_mult(state->a[j + 1], *ptr_sig);
        ptr_sig = cyclic_add(ptr_sig, -1,  ptr_base, L_MEM_SPEECH);
    }
    acc1 += fract_mult(state->a[8],   *ptr_sig);

    ptr_res[288]     = rnd_saturate(acc1 << 3);/*Q31 -> Q15 */
}
