# ----------------------------------------------------------------------------
# Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
# ON Semiconductor), All Rights Reserved
#
# This code is the property of ON Semiconductor and may not be redistributed
# in any form without prior written permission from ON Semiconductor.
# The terms of use and warranty for this code are covered by contractual
# agreements between ON Semiconductor and the licensee.
#
# This is Reusable Code.
#
# -----------------------------------------------------------------------------
#

class baseFileWrapper(object):
    '''
    Base class defining a wrapper for a file
    '''
    
    def __init__(self, wrappedFile):
        ''' Constructir takes the file to be wrapped '''
        self.wrappedFile = wrappedFile

    def read(self, numBytes):
        ''' read must be implemented by a sub-class '''
        raise NotImplementedError("Base class implementation called, must be overridden")

    def write(self, byteBuffer):
        ''' write must be implemented by a sub-class '''
        raise NotImplementedError("Base class implementation called, must be overridden")
            
    def seek(self, offset, fromWhat = 0):
        ''' seek must be implemented by a sub-class '''
        raise NotImplementedError("Base class implementation called, must be overridden")
    
    def tell(self):
        ''' tell must be implemented by a sub-class '''
        raise NotImplementedError("Base class implementation called, must be overridden")

    def close(self):
        ''' close just delegates to the wrapped file '''
        return self.wrappedFile.close()

    @property
    def isReadable(self):
        ''' report if the file is readable ''' 
        return "r" in self.wrappedFile.mode() 

    @property
    def closed(self):
        ''' report ifthe file is closed '''
        return self.wrappedFile.closed
        
